import logging
from decimal import Decimal

NUMBER_TYPES = (int, float, Decimal)

__version__ = "0.0.2"
__name__ = "imagery-mapilio"
__major__ = False
__minor__ = False
__patch__ = True

logger = logging.getLogger('imagery')





import os
import numpy as np
import cv2

from helper.generator import Generator
from imagery.postprocessor import PostProcessor

class Operation:

    @staticmethod
    def image_write(outPath: str, path: str, image: np.ndarray):
        cv2.imwrite(os.path.join(outPath, os.path.basename(path)), image, [cv2.IMWRITE_JPEG_QUALITY, 80])

    @staticmethod
    def image_action(image: np.ndarray, finalBox: list, class_name: str, writePaths: str,
                     rgb_mask: list, copyImage: np.ndarray, config: dict):
        """

        :param image:
        :param finalBox: detected object bounding box
        :param class_name: detected objects category name
        :param writePaths: where will it write
        :param rgb_mask:
        :param copyImage:
        :param config:
        :return: write local directory and detected objects as a array
        """
        xmin, ymin, xmax, ymax = [int(item) for item in finalBox]

        detectedObject = copyImage[ymin:ymax, xmin:xmax]
        imageNumber = os.path.basename(writePaths[1]).split(".")[0]
        imageSaveType = os.path.basename(writePaths[1]).split(".")[1]
        # saveName = imageNumber + "_{}.".format(writePaths[2]) + imagesaveType
        saveName = imageNumber + "_{}.".format(Generator.unique_matchId_generator()) + imageSaveType
        detectedObjectPath = os.path.join("Exports", "ObjectsImage", saveName)
        cv2.imwrite(detectedObjectPath, detectedObject)

        if config.processedimageWrite:
            cX = int((xmin + xmax) / 2)
            cY = int((ymin + ymax) / 2)

            cv2.putText(image, class_name, (cX, cY),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 2)

            if rgb_mask:
                image = np.array(image)
                image = PostProcessor.image_coloring(image, rgb_mask)

        del copyImage

        return detectedObjectPath, image

    @staticmethod
    def image_read_rgb_as_hsv(detectedObjectPath: str) -> np.ndarray:
        return cv2.imread(detectedObjectPath, cv2.COLOR_RGB2HSV)


import cv2
from helper.convertor import Convertor
from typing import Tuple
import numpy as np
from addict import Dict

class PreProcessor:

    @staticmethod
    def getImage(**kwargs) \
            -> Tuple[np.ndarray, int, int, np.ndarray, str, dict, int, int]:
        """
        :param kwargs : |
                    :param path str: image path
                    :param cfg dict: configuration settings
                    :param devicev str:
                    :param remoteImage bool:
        :return: image and its derivatives values such as height, width
        """
        params = Dict(kwargs)

        print('RemoteImage', params.remoteImage)
        if not params.remoteImage:
            image = cv2.imread(params.path, -1)

        else:
            image = Convertor.url_to_image(params.path)

        width = image.shape[1]
        height = image.shape[0]
        if params.device == "web":
            if params.cfg.image.cropImage:
                cropImage = image[params.cfg.image.horizon:height - params.cfg.image.bottomBegin,
                            params.cfg.image.leftBegin:width - params.cfg.image.rightBegin].copy()
            return image, height, width, cropImage, params.path, params.cfg, \
                   params.cfg.split.numrows, params.cfg.split.numcols
        elif params.device == "mobile":
            # TODO it will be dynamic operation
            numrows = round(width / params.cfg.image.mobileTile)
            numcols = round(height / params.cfg.image.mobileTile)
            if not (numcols >=1 and numcols >=1):
                assert "Check your mobileTile parameter"
            cropImage = image.copy()
            return image, height, width, cropImage, params.path, params.cfg, numrows, numcols

    @staticmethod
    def apply_filter(ccfg: dict, image: np.ndarray, *argv) -> np.ndarray:
        """

        :param ccfg:
        :param image:
        :param argv:
        :return:
        """
        for arg in argv:

            if arg == 'canny':
                gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
                medimg = cv2.GaussianBlur(gray, ccfg.image_preprocess.gaussianblur.ksize,
                                          ccfg.image_preprocess.canny.sigmaX)
                canny_edged = cv2.Canny(medimg, ccfg.image_preprocess.canny.threshold1,
                                        ccfg.image_preprocess.canny.threshold2)  # Canny(gray, 50, 100)
                return canny_edged
            elif arg == 'gaussianblur':
                medimg = cv2.GaussianBlur(image, (7, 7), 0)
                return medimg
            elif arg == 'medianblur':
                medblurimg = cv2.medianBlur(image, ccfg.image_preprocess.medianblur.ksize, )
                return medblurimg
            else:
                return image


import numpy as np
from calculation.pixel import Pixel


class Reflection:

    @staticmethod
    def mask_convert_big_image(y0, x0, rgb_mask, horizon):
        mask_covnert_x = []
        mask_covnert_y = []
        rgb_mask = (np.nonzero(rgb_mask))
        if len(rgb_mask[0]) == len(rgb_mask[1]) and rgb_mask is not None:
            for i in range(len(rgb_mask[0])):
                [x, y] = rgb_mask[0][i], rgb_mask[1][i]
                temp = Pixel.calc_origin(y0, x0, [y, x], 1, 1, horizon)
                mask_covnert_x.append(temp[0])
                mask_covnert_y.append(temp[1])
        else:
            raise Exception("Mask points X must be equal Y")

        return [mask_covnert_x, mask_covnert_y, rgb_mask[2]]


"""

Mapilio Image-tools is a python library for some image, prediction functions


"""

from .operations import Operation
from .postprocessor import PostProcessor
from .preprocess import PreProcessor
from .reflection import Reflection

# from .util import __version__, __name__, __major__, __minor__, __patch__


import numpy as np
import cv2
import random

from calculation.pixel import Pixel

class PostProcessor:

    @staticmethod
    def image_coloring(image: np.ndarray, rgb_mask: list) -> np.ndarray:
        """

        :param image: it have processed image
        :param rgb_mask: predicted mask array
        :return:
        """
        for i in range(len(rgb_mask[0])):
            image[rgb_mask[1][i], rgb_mask[0][i], 0] = 5  # y and x coordinates coloring
            image[rgb_mask[1][i], rgb_mask[0][i], 2] = 200  # y and x coordinates coloring

        return image

    @staticmethod
    def vis_bbox(image: np.ndarray, bbox: list, **kwargs) -> np.ndarray:

        """
            Draw bbox,info box on image
        :param image: images to be processed
        :param bbox: [xmin, ymin, xmax, ymax]
        :param kwargs: classname,box,args (string) :custom box for value
        :return: box drawn image
        """

        overlay = image.copy()
        alpha = 0.4  # Transparency factor.
        pixel = 0
        pixel_height = 17
        header_list = []
        header_pix_x = 4
        header_pix_y = 13
        for key, value in kwargs.items():
            text = str(key).capitalize() + ' : ' + str(value)
            x = bbox[0]
            x2 = bbox[2]
            if key == 'classname':
                y, y2 = bbox[1], bbox[1] - 30
                color = (0, 0, 200)
                header_value = (text, (x + header_pix_x, y - header_pix_y))
                header_list.append((text, (x + header_pix_x, y - header_pix_y)))
            elif key == 'box':
                if value:
                    y, y2 = bbox[1], bbox[3]
                    color = (250, 100, 0)
                else:
                    continue
            else:
                y, y2 = bbox[3] + pixel, bbox[3] + pixel_height
                color = (0, 200, 0)
                header_value = (text, (x + header_pix_x, y + header_pix_y + pixel))
                pixel += 8

            header_list.append(header_value)
            cv2.rectangle(overlay, (x, y), (x2, y2), color, -1)
            image_new = cv2.addWeighted(overlay, alpha, image, 1 - alpha, 0)
            pixel_height += 11

        for i in range(len(header_list)):
            cv2.putText(image_new, header_list[i][0], header_list[i][1], cv2.FONT_HERSHEY_DUPLEX, 0.4, (0, 0, 0), 1)

        return image_new

    @staticmethod
    def get_coloured_mask(mask: list) -> list:
        """
        random_colour_masks
          parameters:
            - image - predicted masks
          method:
            - the masks of each predicted object is given random colour for visualization
        """
        colours = [[0, 255, 0], [0, 0, 255], [255, 0, 0], [0, 255, 255], [255, 255, 0], [255, 0, 255], [80, 70, 180],
                   [250, 80, 190], [245, 145, 50], [70, 150, 250], [50, 190, 190]]
        r = np.zeros_like(mask).astype(np.uint8)
        g = np.zeros_like(mask).astype(np.uint8)
        b = np.zeros_like(mask).astype(np.uint8)
        r[mask == 1], g[mask == 1], b[mask == 1] = colours[random.randrange(0, 10)]
        coloured_mask = np.stack([r, g, b], axis=2)
        return coloured_mask

    def get_mask_polygon_to_rectangle(mask: list) -> list:
        """

        :return:
        """
        maskx = np.any(mask, axis=0)
        masky = np.any(mask, axis=1)
        xmin = np.argmax(maskx)
        ymin = np.argmax(masky)
        xmax = len(maskx) - np.argmax(maskx[::-1])
        ymax = len(masky) - np.argmax(masky[::-1])

        return xmin, ymin, xmax, ymax

    @staticmethod
    def get_segmentation(mask: list, x0: int, y0: int, horizon: int, image: np.ndarray) -> np.ndarray:
        """

        :param mask:
        :param x0:
        :param y0:
        :param horizon:
        :param image:
        :return:
        """
        maskRaw = mask[0]
        mask_ = maskRaw.mul(255).byte().cpu().numpy()
        contours, _ = cv2.findContours(
            mask_.copy(), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

        if contours:
            contour = np.flip(contours[0], axis=1)
            contour = contour.ravel().tolist()
            segmentation = Pixel.calc_origin(y0, x0, contour, 1, 1, horizon)
            cv2.drawContours(image, segmentation, -1, (255, 0, 0), 2, cv2.LINE_AA)

        else:
            segmentation = None

        return segmentation

    @staticmethod
    def get_histogram_matcher(detect_1: np.ndarray, detect_2: np.ndarray) -> float:
        # The histogram size refers to the number of bins in the histogram.
        h_bins = 50
        s_bins = 60
        histSize = [h_bins, s_bins]
        # hue varies from 0 to 179, saturation from 0 to 255
        h_ranges = [0, 180]
        s_ranges = [0, 256]
        ranges = h_ranges + s_ranges  # concat lists
        # Use the 0-th and 1-st channels
        channels = [0, 1]

        # Calculate the Hist for each images
        # Calculate the histogram and normalize it
        hist_img1 = cv2.calcHist([detect_1], channels, None, histSize, ranges, accumulate=False)
        cv2.normalize(hist_img1, hist_img1, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)
        hist_img2 = cv2.calcHist([detect_2], channels, None, histSize, ranges, accumulate=False)
        cv2.normalize(hist_img2, hist_img2, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX)

        # find the metric value
        metric_val = cv2.compareHist(hist_img1, hist_img2, cv2.HISTCMP_BHATTACHARYYA)
        del hist_img1, hist_img2

        return metric_val
